#!/usr/bin/env python

from abc import ABC, abstractmethod
from typing import Any, Dict, List, NamedTuple

import jsonschema
import jsonschema.exceptions
from importlib_metadata import entry_points
from referencing.jsonschema import Schema

from jiav import exceptions


class Result(NamedTuple):
    successful: bool
    output: List[str]
    errors: List[str]


class BaseBackend(ABC):
    """
    jiav base backend object

    All backends should inherit from this object

    Attributes:
        name   - Backend name
        valid  - Whether backend is valid
        schema - json_schema to be used to verify that the supplied setp
                 is valid according to the backends's requirments
        step   - Instructions to perform according to backend
    """

    # SCHEMA: Dict = {}

    def __init__(self, name: str, schema: Schema, step: Dict) -> None:
        self.name = name if name else "placeholder"
        self.schema = schema
        self.step = step
        self.result: Result

    def validate_schema(self) -> bool:
        """
        Validates that the verification step is obeying the required
        schema

        Returns True if backend schema is valid
        """
        try:
            jsonschema.validate(instance=self.step, schema=self.schema)
        except jsonschema.exceptions.ValidationError as e:
            raise exceptions.InvalidManifestException(e)
        return True

    @abstractmethod
    def execute_backend(self) -> None:
        """
        Executes backend

        This function should be overridden by each individual backend

        Returns False if this method is invoked directly and not overriden
        by backend
        """
        pass


def import_backends() -> Dict[str, Dict[str, Any]]:
    """
    Imports all valid backends from jiav/api/backends directory

    All valid backends are inherited from BaseBackend object

    Returns:
        exposed_backends - List of valid backend objects based on inherited
        from BaseBackend object
    """
    discovered_backends = entry_points(group="jiav.backend")
    exposed_backends: Dict[str, Dict[str, Any]] = {}
    for backend in discovered_backends:
        exposed_backends.update(
            {backend.name: {"version": backend.dist.version, "class": backend.value}}
        )
    return exposed_backends



#!/usr/bin/env python

from typing import List, Union

from jira import Issue
from jiav import logger
from jiav.backend import BaseBackend, Result
from jiav.jira import JiraConnection

jiav_logger = logger.subscribe_to_logger()


class JiraIssueBackend(BaseBackend):
    """
    JiraIssueBackend object

    Checks that a Jira issue is in a desired status

    Attributes:
        name   - Backend name
        schema - json_schema to be used to verify that the supplied step is
                 valid according to the backends's requirements
        step   - Backend excution instructions
    """

    MOCK_STEP = {"issue": "TEST-1", "status": "Done"}
    SCHEMA = {
        "type": "object",
        "required": ["issue", "status"],
        "properties": {"issue": {"type": "string"}, "status": {"type": "string"}},
        "additionalProperties": False,
    }

    def __init__(self) -> None:
        self.name = "jira_issue"
        self.schema = self.SCHEMA
        self.step = self.MOCK_STEP
        super().__init__(name=self.name, schema=self.schema, step=self.step)

    # Overrdie method of BaseBackend
    def execute_backend(self) -> None:
        """
        Execute backend

        Returns a namedtuple describing the jiav manifest execution
        """
        # Parse required arugments
        issue: Union[Issue, str] = self.step["issue"]
        issue_status: str = self.step["status"]
        output: List = []
        errors: List = []
        successful: bool = False
        # Reusing the original JiraConnection object since the class is a singleton
        jira_connection: JiraConnection = JiraConnection()  # type: ignore
        remote_issue: Union[Issue, None] = None
        remote_issue_status: str = ""
        jiav_logger.debug(f"Issue: {issue}")
        jiav_logger.debug(f"Status: {issue_status}")
        try:
            remote_issue = jira_connection.jira.issue(id=issue)
            remote_issue_status = str(remote_issue.get_field("status"))
            if remote_issue_status != issue_status:
                jiav_logger.error(
                    " ".join(
                        [
                            f"Issue '{issue}' status '{remote_issue_status}' does",
                            f"not match the desired status '{issue_status}'",
                        ]
                    )
                )
                errors.append(
                    (
                        f"Issue '{issue}' status '{remote_issue_status}' does",
                        f"not match the desired status '{issue_status}'",
                    )
                )
            else:
                successful = True
                output.append(
                    f"Jira issue '{issue}' is in the desired status '{issue_status}'"
                )
        except Exception as e:
            jiav_logger.error(e.text)  # type: ignore # pyright: ignore # pylint: disable=E1101 # noqa: E501
            errors.append(e.text)  # type: ignore # pyright: ignore # pylint: disable=E1101 # noqa: E501
        self.result = Result(successful, output, errors)



#!/usr/bin/env python

import os
from typing import List

from jiav import logger
from jiav.backend import BaseBackend, Result

# Subscribe to logger
jiav_logger = logger.subscribe_to_logger()


class LineInFileBackend(BaseBackend):
    """
    LineInFile backend object

    Checks if line is in file

    Attributes:
        name   - Backend name
        schema - json_schema to be used to verify that the supplied step is
        valid according to the backends's requirements
        step   - Backend excution instructions
    """

    MOCK_STEP = {"path": "/path/to/file", "line": "line_in_file"}
    SCHEMA = {
        "type": "object",
        "required": ["path", "line"],
        "properties": {"path": {"type": "string"}, "line": {"type": "string"}},
        "additionalProperties": False,
    }

    def __init__(self) -> None:
        self.name = "lineinfile"
        self.schema = self.SCHEMA
        self.step = self.MOCK_STEP
        super().__init__(name=self.name, schema=self.schema, step=self.step)

    # Overrdie method of BaseBackend
    def execute_backend(self) -> None:
        """
        Execute backend

        Returns a namedtuple describing the jiav manifest execution
        """
        # Parse required arugments
        file: str = self.step["path"]
        line: str = self.step["line"]
        output: List[str] = []
        errors: List[str] = []
        successful = False
        jiav_logger.debug(f"File: '{file}'")
        jiav_logger.debug(f"Line: '{line}'")
        # Check if file exits
        if not os.path.exists(file):
            errors.append(f"File '{file}' does not exist")
            jiav_logger.error(f"File '{file}' does not exist")
        else:
            try:
                with open(file) as f:
                    file_content = f.read()
                    if line not in file_content:
                        errors.append(f"Line '{line}' is not present in file '{file}'")
                        jiav_logger.error(
                            f"Line '{line}' is not present in file '{file}'"
                        )
                    else:
                        successful = True
                        output.append(f"Line '{line}' found in '{file}'")
                        jiav_logger.debug(f"Line '{line}' found in '{file}'")
            except Exception as e:
                errors.append(f"OS exception: {str(e)}")
        self.result = Result(successful, output, errors)



#!/usr/bin/env python

import json
from typing import Any, Dict, List

import yaml
from jira import Issue
from rich import box, print_json
from rich.console import Console
from rich.table import Table


def prepare_summary(issues: List[Issue] = [], format: str = "table") -> None:
    """
    Prepares a summary or verified issues

    Arguments:
        issues - Jira issues

        format - Output format
    """
    if format == "json":
        format_json("verified_issues", issues)
    elif format == "yaml":
        format_yaml("verified_issues", issues)
    else:
        format_table("Verified Issues", issues)


def construct_dict(title: str, issues: List[Issue]) -> Dict[Any, Any]:
    """
    Constructs a dictionary from received items

    Arguments:
        title - Name of the key

        issues - Jira issues that will populate the value
    """
    dictionary: dict = dict({title.lower(): list()})
    for issue in issues:
        issue_assignee = (
            issue.fields.assignee.displayName if issue.fields.assignee else "Unassigned"
        )
        dictionary[title].append(
            dict(
                {
                    "issue": issue.key,
                    "summary": issue.fields.summary,
                    "status": issue.fields.status.name,
                    "assignee": issue_assignee,
                    "reporter": issue.fields.reporter.displayName,
                    "comments": issue.fields.comment.total,  # type: ignore
                }
            )
        )
    return dictionary


def format_json(title: str, issues: List[Issue]) -> None:
    """
    Prints a JSON of issues

    Arguments:
        title  - Name of the key

        issues - List of issues
    """
    j = json.dumps(construct_dict(title, issues))
    print_json(json=j)


def format_yaml(title: str, issues: list) -> None:
    """
    Prints a YAML of issues

    Arguments:
        title  - Name of the key

        issues - List of issues
    """
    print(
        yaml.dump(
            construct_dict(title, issues),
            explicit_start=False,
            default_flow_style=False,
        )
    )


def format_table(title: str, issues: List[Issue]) -> None:
    """
    Prints a table of issues

    Arguments:
        title  - Title of the table

        issues - List of issues
    """
    table_columns: List[str] = [
        "Issue",
        "Summary",
        "Status",
        "Assignee",
        "Reporter",
        "Comments",
    ]
    table = Table(title=title, box=box.SIMPLE_HEAD, *table_columns)
    for issue in issues:
        issue_assignee = (
            issue.fields.assignee.displayName if issue.fields.assignee else "Unassigned"
        )
        table.add_row(
            str(issue),
            issue.fields.summary,
            str(issue.fields.status),
            (f"{issue_assignee} "),
            (f"{issue.fields.reporter.displayName} "),
            str(issue.fields.comment.total),  # type: ignore
        )
    Console().print(table)



#!/usr/bin/env python

from __future__ import annotations

from abc import ABC
from typing import Any, List, Union

from jira import JIRA, Issue, JIRAError
from requests.exceptions import ConnectionError

from jiav import exceptions, logger

jiav_logger = logger.subscribe_to_logger()


class JiraConnection(ABC):
    """
    Jira connection

    Attributes:
        jira - jira.JIRA object if authenticated successfully.
    """

    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "JiraConnection":
        if not cls._instance:
            cls._instance = super(JiraConnection, cls).__new__(cls)
        return cls._instance

    def __init__(self, url: str, username: str, access_token: str):
        """
        Attempts to authenticate with Jira API

        Arguments:
            url           - Jira URL
            instance_type - Instance type (cloud or hosted)
            access_token  - Personal Access Token to authenticate with
        """
        # Initiall connection is used to discover details about the Jira instance
        if not hasattr(self, "jira"):
            try:
                self.jira = JIRA(server=url)
            except JIRAError as e:
                if "JiraError HTTP 404 url" in str(e):
                    raise exceptions.NoJiraRestAPIEndpoint(url) from None
                else:
                    raise exceptions.JiraUnhandledException()
            # NOTE: Failed to contact the server enitrely, perhaps should
            #       raise a unique exception for this case
            except ConnectionError:
                raise exceptions.NoJiraRestAPIEndpoint(url) from None
            instance_type = self.jira.deploymentType
            # Authenticate with a Jira cloud instance
            if instance_type == "Cloud":
                if not username:
                    raise exceptions.JiraMissingCredentials("Username") from None
                self.jira = JIRA(server=url, basic_auth=(username, access_token))
            # Authenticate with a self-hosted Jira instance
            elif instance_type == "Server":
                if username:
                    jiav_logger.warning(
                        "Username argument is omitted for self-hosted Jira instances"
                    )
                self.jira = JIRA(server=url, token_auth=access_token)
            # Check if authentication was successful
            try:
                self.jira.myself()
            except JIRAError as e:
                if "JiraError HTTP 401 url" in str(e):
                    raise exceptions.JiraAuthenticationFailed(url) from None
                else:
                    raise exceptions.JiraUnhandledException()
            jiav_logger.info(
                " ".join(
                    [
                        f"Successfully authenticated with Jira instance '{url}' of",
                        f"type '{instance_type}'",
                    ]
                )
            )

    def fetch_issues(self, issues: List[str], jql: str) -> List[Issue]:
        """
        Attempt to fetch issues from Jira
        Requires jira attribute to contain jira.JIRA
        (successful authentication using authenticate method)

        Arguments:
            issues - List of Jira issues
            jql    - Jira Query Language query

        Returns:
            remote_issues - Fetched Jira issues
        """
        remote_issues: List[Issue] = []

        # Iterate over provided issues
        if issues:
            for issue in issues:
                try:
                    remote_issue = self.jira.issue(issue)
                    remote_issues.append(remote_issue)
                except JIRAError as e:
                    if "You do not have the permission to see " in e.text:
                        jiav_logger.error(
                            f"You do not have permissions to view {issue}"
                        )
                    # NOTE: In Jira cloud instances, the returned output is
                    #       'Issue does not exist',
                    #       While self-hosted instances, the otuput is:
                    #       'Issue Does Not Exist'
                    elif "Issue does not exist".lower() in e.text.lower():
                        jiav_logger.error(f"Issue '{issue}' does not exist")
                    else:
                        raise exceptions.JiraUnhandledException()
        # Perform JQL query
        elif jql:
            try:
                remote_issues = self.jira.search_issues(jql, json_result=False)  # type: ignore # noqa: E501
            except JIRAError as e:
                if "Error in the JQL Query:" in e.text:
                    raise exceptions.JQLError(jql, e.text) from None
                elif "for field" and "is invalid" in e.text:
                    raise exceptions.InvalidKeyInJQL(jql, e.text) from None
                else:
                    raise exceptions.JiraUnhandledException()
            if not remote_issues:
                raise exceptions.JQLReturnedNothing(jql)

        if not remote_issues:
            raise exceptions.NoIssuesFound()
        else:
            jiav_logger.info(f"Discovered issues: {remote_issues}")
        return remote_issues

    def check_if_status_is_valid(
        self, issue: Issue, desired_status: str
    ) -> Union[None, str]:
        """
        Checks if desired status is a valid status for this issue

        Arguments:
            issue          - Jira issue
            desired_status - Desired status requested by jiav manifest

        Returns:
            transition_id - Workflow transition ID to use when updating
                            the status of the issue
        """
        transition_id: Union[None, str] = None
        issue_transitions = self.jira.transitions(issue)
        for status in issue_transitions:
            if desired_status == status["name"]:
                transition_id = status["id"]
        return transition_id

    def upload_attachment(self, issue: Issue, file_path: str) -> None:
        """
        Uploads an attachment to issue

        Arguments:
            issue    - Jira issue
            filename - Path of the file to add as an attachment
        """
        try:
            self.jira.add_attachment(issue=issue, attachment=file_path)
            jiav_logger.info(f"Uploaded attachment to issue '{issue}'")
        except JIRAError:
            raise exceptions.JiraUnhandledException()

    def update_issue_status(self, issue: Issue, transition_id: str) -> None:
        """
        Update issue status using transition ID discovered by
        jiav.utils.jira.JiraConnection.check_if_status_is_valid

        Arguments:
            issue         - Jira issue
            transition_id - Workflow transition ID to use when updating
                            the status of the issue
        """
        try:
            self.jira.transition_issue(issue, transition=transition_id)
            issue.update()
            jiav_logger.info(f"Updated status for '{issue}'")
        except JIRAError:
            raise exceptions.JiraUnhandledException()

    def post_comment(self, issue: Issue, comment: str) -> None:
        """
        Posts a comment with the result of verification process

        Arguments:
            issue   - Jira issue
            comment - Comment to post
        """
        try:
            self.jira.add_comment(issue, comment)
            jiav_logger.info(f"Posted a comment in '{issue}'")
        except JIRAError:
            raise exceptions.JiraUnhandledException()



#!/usr/bin/env python

import importlib
from abc import ABC
from typing import Any, Dict, List, Type, Union

import jsonschema
import jsonschema.exceptions
import yaml
import yaml.composer
import yaml.parser
import yaml.scanner

from jiav import exceptions, logger
from jiav.backend import BaseBackend, import_backends

jiav_logger = logger.subscribe_to_logger()


class Manifest(ABC):
    """
    A jiav Manifest

    Attributes:
        manifest_text      - Manifest text fetched from comment
        successful         - Result of manifest execution
        verified_status    - Jira status to transition to
        verification_steps - Verification steps to perform
        backend_steps      - Parsed backend steps to execute
    """

    # A JSON schema for jiav manifest
    ROOT_JIAV_SCHEMA = {
        "type": "object",
        "required": ["jiav"],
        "properties": {
            "jiav": {
                "type": "object",
                "required": ["verified_status", "verification_steps"],
                "properties": {
                    "verification_steps": {"type": "array"},
                    "verified_status": {"type": "string"},
                },
            }
        },
    }
    # A JSON schema for individual steps
    STEP_SCHEMA = {
        "type": "object",
        "required": ["name", "backend"],
        "properties": {"name": {"type": "string"}, "backend": {"type": "string"}},
    }

    def __init__(self, manifest_text: str) -> None:
        self.manifest: Dict = self.validiate_text_contains_manifest(manifest_text)
        self.successful: bool = False
        self.verified_status: str = self.manifest["jiav"]["verified_status"]
        self.verification_steps: List[Dict[str, str]] = self.manifest["jiav"][
            "verification_steps"
        ]
        self.backend_steps: List[BaseBackend] = []
        self.validate_verifications_steps()

    def validiate_text_contains_manifest(self, text: str) -> Dict[Any, Any]:
        """
        Performs an initial validation of text, checks if text is a valid
        manifest
        """
        # Attempt to parse YAML from comment
        m: Dict[Any, Any] = {}
        try:
            m = yaml.safe_load(text)
        except yaml.scanner.ScannerError:
            jiav_logger.debug("Comment can not be parsed as YAML")
        except yaml.parser.ParserError:
            jiav_logger.debug("Comment can not be parsed as YAML")
        except yaml.composer.ComposerError:
            jiav_logger.debug("Comment can not be parsed as YAML")
        # Attempt to validate YAML according to schema
        try:
            jsonschema.validate(instance=m, schema=self.ROOT_JIAV_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise exceptions.InvalidManifestException(e) from None
        return m

    def backend_is_valid(self, backend: str) -> bool:
        """
        Validates requested backend

        Params:
            backend - Requested backend

        Returns a boolean value if the requested backed is installed
        """
        return backend in import_backends().keys()

    def validate_verification_step(self, step: Dict[str, Any]) -> bool:
        """
        Validates a verification step

        Arguments:
            step - Verification step

        Returns True if successfully validated verification step
        """
        backend_name: str = step["backend"]
        backend_class_name: str = import_backends()[backend_name]["class"]
        backend_module = importlib.import_module(backend_class_name.split(".")[0])
        backend_module_class: str = backend_class_name.split(".")[1]
        backend_class: Type[BaseBackend] = getattr(backend_module, backend_module_class)
        # Construct initial class
        backend_instance = backend_class()  # type: ignore
        # Remove unnecessary keys for schema validation
        del step["backend"]
        del step["name"]
        backend_instance.step = step
        # Validate backend schema
        backend_instance.validate_schema()
        self.backend_steps.append(backend_instance)
        return True

    def execute_manifest(self) -> None:
        """
        Executes manifest

        Returns:
            execution_output - jiav manifest execution output
        """
        # Init variables
        execution_output: List[Union[str, List[str]]] = []
        # Iterate over steps in manifest
        for step in self.backend_steps:
            jiav_logger.debug("Executing backend '{}'".format(step.name))
            execution_output.append("Output of backend '{}':\n".format(step.name))
            try:
                step.execute_backend()
            except Exception:
                raise exceptions.BackendExecutionFailed(step.name)
            if step.result.successful:
                execution_output.append(step.result.output)
            # If one of the steps failed, immediately fail and don't continue
            else:
                # If stderr contains output, log it
                if step.result.errors:
                    execution_output.append(step.result.errors)
                else:
                    execution_output.append(step.result.output)
                break
            self.successful = True
        self.execution_output = execution_output
        if self.successful:
            jiav_logger.info("Manifest was executed successfully")
        else:
            jiav_logger.error("Manifest has failed to execute")

    def validate_verifications_steps(self) -> None:
        """
        Validates verification_steps from jiav manifest
        """
        steps = self.verification_steps
        # Iterate over steps in manifest
        for step in steps:
            step_backend = step.get("backend")
            # Attempt to validate step according to schema
            try:
                jsonschema.validate(instance=step, schema=self.STEP_SCHEMA)
            except jsonschema.exceptions.ValidationError as e:
                raise exceptions.InvalidManifestException(e)

            if not self.backend_is_valid(str(step_backend)):
                raise exceptions.InvalidBackend(str(step_backend))
            self.validate_verification_step(step)



#!/usr/bin/env python

import os
import re

from jiav import logger
from jiav.backend import BaseBackend, Result

MOCK_STEP = {"path": "/path/to/file", "regex": r"^.*$"}
SCHEMA = {
    "type": "object",
    "required": ["path", "regex"],
    "properties": {"path": {"type": "string"}, "regex": {"type": "string"}},
    "additionalProperties": False,
}

# Subscribe to logger
jiav_logger = logger.subscribe_to_logger()


class RegexInFileBackend(BaseBackend):
    """
    RegexInFile backend object

    Checks if regex is in file

    Attributes:
        name   - Backend name
        schema - json_schema to be used to verify that the supplied step is
        valid according to the backends's requirements
        step   - Backend excution instructions
    """

    def __init__(self) -> None:
        self.name = "regexinfile"
        self.schema = SCHEMA
        self.step = MOCK_STEP
        super().__init__(name=self.name, schema=self.schema, step=self.step)

    # Override method of BaseBackend
    def execute_backend(self) -> None:
        """
        Execute backend

        Returns a namedtuple describing the jiav manifest execution
        """
        # Parse required arugments
        file = self.step["path"]
        regex = self.step["regex"]
        output = list()
        errors = list()
        successful = False
        compiled_regex = re.compile(rf"{regex}")
        if not os.path.exists(file):
            errors.append(f"File {file} does not exist")
        else:
            try:
                with open(file) as f:
                    file_content = f.read()
                    if not compiled_regex.search(file_content):
                        errors.append(
                            f"Regex '{regex}' is not present in file '{file}'"
                        )
                        jiav_logger.error(
                            f"Regex '{regex}' is not present in file '{file}'"
                        )
                    else:
                        successful = True
                        output.append(f"Regex '{regex}' found in '{file}'")
                        jiav_logger.debug(f"Regex '{regex}' found in '{file}'")
            except Exception as e:
                errors.append(f"OS exception: {str(e)}")
        self.result = Result(successful, output, errors)



#!/usr/bin/env python

from jsonschema.exceptions import ValidationError


class jiavException(Exception):
    """
    Base jiav exception
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class NoJiraRestAPIEndpoint(jiavException):
    """
    Raised when the provided URL is not a valid Jira Rest API endpoint
    """

    def __init__(self, url: str) -> None:
        message = f"Provided url '{url}' does not expose a Jira Rest API endpoint"
        super().__init__(message)


class JiraMissingCredentials(jiavException):
    """
    Raised when user did not provide all required credentials
    """

    def __init__(self, missing_credentials: str) -> None:
        message = f"{missing_credentials} is required for this instance"
        super().__init__(message)


class JiraAuthenticationFailed(jiavException):
    """
    Raised when user did not authenticate correctly with Jira instance
    """

    def __init__(self, url: str) -> None:
        message = (
            f"Failed to authenticate with instance {url}, please check your credentials"
        )
        super().__init__(message)


class PermissionsError(jiavException):
    """
    Raised when failed to fetch content because of permissions
    """

    def __init__(self, issue: str) -> None:
        message = (
            f"You do not have permissions to view {issue}, "
            "please ensure you are using the correct token."
        )
        super().__init__(message)


class IssueNotExists(jiavException):
    """
    Raised when no valid issues were found
    """

    def __init__(self, issue: str) -> None:
        message = f"Issue {issue} does not exist."
        super().__init__(message)


class JQLError(jiavException):
    """
    Raised when JQL error occurs
    """

    def __init__(self, jql: str, err: str) -> None:
        message = f"{err}.\nQuery: {jql}"
        super().__init__(message)


class JQLReturnedNothing(jiavException):
    """
    Raised when JQL returns no issues
    """

    def __init__(self, jql: str) -> None:
        message = (
            "No issues returned by query, please ensure your query is valid"
            f"\nQuery: {jql}"
        )
        super().__init__(message)


class InvalidKeyInJQL(jiavException):
    """
    Raised when an invalid key in JQL
    """

    def __init__(self, jql: str, err: str) -> None:
        message = f"{err}.\nQuery: {jql}"
        super().__init__(message)


class NoIssuesFound(jiavException):
    """
    Raised when no issues were found
    """

    def __init__(self) -> None:
        message = "No issues were found"
        super().__init__(message)


class JiraUnhandledException(jiavException):
    """
    Raised when an unhandled exception is raised by
    requests.exceptions.HTTPError
    """

    def __init__(self) -> None:
        message = "Unhandled excpetion was raised, please report it"
        super().__init__(message)


class InvalidManifestException(jiavException):
    """
    Raised when an invalid manifest is supplied
    """

    def __init__(self, py_err: ValidationError) -> None:
        super().__init__(str(py_err))


class InvalidBackend(jiavException):
    """
    Raised when an unsupported backend is requested
    """

    def __init__(self, backend: str) -> None:
        message = f"'{backend}' is not a supported backend"
        super().__init__(message)


class BackendExecutionFailed(jiavException):
    """
    Raised when an unexxpected error occurs during backend execution
    """

    def __init__(self, backend: str) -> None:
        message = f"'{backend}' execution failed"
        super().__init__(message)



#!/usr/bin/env python

import importlib.metadata
from typing import List, Tuple, Union

from jira import Issue

from jiav import content, exceptions, logger, manifest
from jiav.jira import JiraConnection

jiav_logger = logger.subscribe_to_logger()


def prepare_jiav_comment(
    status: str,
    manifest_comment: str,
    successful: bool = False,
    upload_attachment: bool = False,
) -> str:
    """
    Prepares jiav comment containing the execution result

    Arguments:
        successful       - Result of jiav manifest execution
        status           - Desired status of verified issue
        manifest_comment - Jira comment containing jiav manifest
        attachment       - Attachment to uploaded if required

    Returns:
        generated_comment - Comment that will be posted
    """
    generated_comment: List = []
    text_body: List = []
    # If jiav manifest execution is successful
    text_body.append(f"Executed manifest on comment ID: *{{{{{manifest_comment}}}}}*.")
    if successful:
        text_body.append("Manifest executed successfully.")
        text_body.append(f"Verifying issue as *{{{{{status}}}}}*.")
    # If jiav manifest execution failed
    else:
        text_body.append("Failed to verify issue using provided manifest.")
    if upload_attachment:
        text_body.append(
            "Uploading attachment of execution, *attachment is not private*."
        )
    else:
        text_body.append("Attachment was not uploaded, view logs on executor.")
    generated_comment.append("{panel:title=jiav Report}")
    generated_comment.append("\n".join(text_body))
    generated_comment.append("{panel}")
    generated_comment.append(
        " ".join(
            [
                "Generated by [jiav|https://github.com/vkhitrin/jiav],",
                f"version: {{{{{importlib.metadata.version('jiav')}}}}}.",
            ]
        )
    )
    return "\n".join(generated_comment)


def process_comments(
    comments: List, allow_public_comments: bool = False
) -> Tuple[Union[manifest.Manifest, bool], str]:
    valid_manifest: Union[manifest.Manifest, bool] = False
    jiav_comment: str = ""
    # Iterate over comments in reverse order (from latest to oldest)
    for idx, comment in reversed(list(enumerate(comments))):
        jiav_logger.debug(f"Looking at comment '#{idx}'")
        # Attempt to validate comment according to jiav manifest
        if not allow_public_comments:
            try:
                comment.visibility.type
            except AttributeError:
                jiav_logger.debug(f"Comment {idx} is not private, skipping")
                continue
        try:
            valid_manifest = manifest.Manifest(manifest_text=comment.body)
        except exceptions.InvalidManifestException:
            jiav_logger.debug(f"Comment {idx} does not contain a valid manifest")
        except exceptions.InvalidBackend as e:
            jiav_logger.error(e)
        if valid_manifest:
            jiav_comment = comment
            break
    return valid_manifest, jiav_comment


def verify_issues(
    jira_connection: JiraConnection,
    issues: List[Issue] = [],
    upload_attachment: bool = False,
    allow_public_comments: bool = False,
    no_comment_on_failure: bool = False,
    dry_run: bool = False,
) -> List[Issue]:
    """
    Attempts to verify issues

    Arguments:
        issues - List of issues from jiav.utils.jira.discover_valid_issues function

        jira_connection - Jira connection object defined by
        jiav.utils.jira.JiraConnection

        upload_attacment - Upload attachment if requested

        public_comment_allowed - Allow manifest from public comments

        dry_run - Do not update issues

    Returns:
        verified_issues - List of issues that were successfully verified
    """
    # Init variables
    verified_issues: List[Issue] = []
    # Iterate over issues
    for issue in issues:
        jiav_manifest: Union[manifest.Manifest, bool] = False
        manifest_comment: str = ""
        desired_status: str = ""
        jiav_logger.info(f"Looking at issue '{issue}'")
        jiav_manifest, manifest_comment = process_comments(
            comments=issue.fields.comment.comments,
            allow_public_comments=allow_public_comments,
        )
        # If valid manifest was provided in comments
        if isinstance(jiav_manifest, manifest.Manifest):
            jiav_logger.info(f"Valid manifest was found in issue '{issue}'")
            desired_status = jiav_manifest.verified_status
            current_status: str = issue.fields.status.name
            # If issue is already in the desired status, we skip it
            if desired_status == current_status:
                jiav_logger.info(
                    " ".join(
                        [
                            f"Issue '{issue}' is already in the",
                            f"desired status '{desired_status}'",
                        ]
                    )
                )
                break
            transition_id = jira_connection.check_if_status_is_valid(
                issue=issue, desired_status=desired_status
            )
            if not transition_id:
                jiav_logger.error(
                    " ".join(
                        [
                            f"Desired status '{desired_status}' is not valid in",
                            "this instance or it can not be",
                            "trasnsitioned from current status",
                            f"'{current_status}'",
                        ]
                    )
                )
                break
            # Execute according to jiav request
            try:
                jiav_manifest.execute_manifest()
            except exceptions.BackendExecutionFailed as e:
                jiav_logger.exception(e)
            if not dry_run:
                comment = prepare_jiav_comment(
                    successful=jiav_manifest.successful,
                    status=jiav_manifest.verified_status,
                    manifest_comment=manifest_comment,
                    upload_attachment=upload_attachment,
                )
                # Post comment with the execution when applicable
                if not no_comment_on_failure or jiav_manifest.successful:
                    jira_connection.post_comment(issue=issue, comment=comment)

                # Add attachment if requested by user
                if upload_attachment:
                    # Generate tempfile with the jiav manifest execution output
                    try:
                        temp_file = content.write_content_to_tempfile(
                            jiav_manifest.execution_output
                        )
                        jira_connection.upload_attachment(
                            issue=issue, file_path=temp_file.name
                        )
                    except AttributeError:
                        jiav_logger.error(
                            " ".join(
                                [
                                    "Failed to create attachment, no execution",
                                    "output in manifest execution",
                                ]
                            )
                        )

                # If manifest executed successfully
                if jiav_manifest.successful:
                    jira_connection.update_issue_status(
                        issue=issue, transition_id=transition_id
                    )
                    verified_issues.append(issue)
        else:
            jiav_logger.error(f"Valid manifest was not found in issue '{issue}'")
    return verified_issues



#!/usr/bin/env python

import importlib.metadata
from logging import Logger
from typing import Any, List, Tuple

import rich_click as click
from jira import Issue

from jiav import exceptions, logger, summary, verification
from jiav.backend import import_backends
from jiav.jira import JiraConnection
from jiav.manifest import Manifest

click.rich_click.COLOR_SYSTEM = "truecolor"
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.OPTION_GROUPS = {
    "jiav verify": [
        {
            "name": "Main options",
            "options": ["--jira", "--access-token", "--username"],
        },
        {
            "name": "Issue options",
            "options": ["--issue", "--query"],
        },
        {
            "name": "Dangerous options",
            "options": ["--upload-attachment", "--allow-public-comments"],
        },
    ]
}

jiav_logger: Logger = logger.configure_logger()


class MutuallyExclusiveOption(click.Option):
    def __init__(self, *args: Any, **kwargs: Any):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            opt = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with"
                " arguments: [" + opt + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(
        self, ctx: click.Context, opts: Any, args: Any
    ) -> Tuple[Any, Any]:
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


def _version_option() -> str:
    """
    Build the string printed by jiav --version
    """
    jiav_version = importlib.metadata.version("jiav")
    backend_output = "\n".join(
        [
            f"  - {name}, version {version}"
            for name, version in import_backends().items()
        ]
    )

    output = (
        f"jiav, version {jiav_version}\n\n" "Installed Backends:\n" f"{backend_output}"
    )

    return output


@click.group(context_settings={"auto_envvar_prefix": "JIAV"})
@click.version_option(package_name="jiav", message=_version_option())
def jiav() -> None:
    pass


@click.command()
@click.option(
    "-f",
    "--from-file",
    type=click.File("r"),
    help="Path to local file to validate.",
    required=True,
)
def validate_manifest(from_file: click.File) -> None:
    """Validate manifest locally."""
    content = from_file.read()  # type: ignore
    jiav_logger.info(f"File content:\n{content}")
    try:
        Manifest(manifest_text=content)
        jiav_logger.info("Provided YAML is valid")
    except Exception as e:
        jiav_logger.error(e)
        jiav_logger.error("Provided manifest is not valid")
        raise SystemExit(5)


@click.command()
@click.option("-j", "--jira", type=str, help="Jira URL.", required=True)
@click.option(
    "-a",
    "--access-token",
    type=str,
    help=(
        " ".join(
            [
                "Personal Access Token (PAT) for self-hosted instances or ",
                "an API token for cloud instances.",
            ]
        )
    ),
    required=True,
)
@click.option(
    "-u",
    "--username",
    type=str,
    help="Cloud Jira username NOTE: Not required for self-hosted instances.",
)
@click.option(
    "-i",
    "--issue",
    help="Issue to verify.",
    multiple=True,
    type=str,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["query"],
)
@click.option(
    "-q",
    "--query",
    help="JQL query.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["issue"],
)
@click.option(
    "--upload-attachment",
    help=(
        " ".join(
            [
                "Uploads attachment of execution, this is not safe since all",
                "users who can access the ticket will be able to view it;",
                "refer to https://jira.atlassian.com/browse/JRASERVER-3893",
            ]
        )
    ),
    flag_value=True,
)
@click.option(
    "--allow-public-comments",
    help=(
        " ".join(
            [
                "Allows to read manifest from non-private comments; this is",
                "potentially dangerous since unexpected users will be able to",
                "provide a manifest.",
            ]
        )
    ),
    flag_value=True,
)
@click.option(
    "--no-comment-on-failure",
    help="Do not post a comment on failed manifest execution.",
    flag_value=True,
)
@click.option(
    "--dry-run",
    help="Execute manifest without updating issues.",
    flag_value=True,
)
@click.option(
    "--debug",
    help="Enable debug logging.",
    flag_value=True,
)
@click.option(
    "--format",
    type=click.Choice(["table", "json", "yaml"], case_sensitive=False),
    help="Output format.",
)
def verify(
    jira: str,
    access_token: str,
    username: str,
    issue: List[str],
    query: str,
    upload_attachment: bool,
    allow_public_comments: bool,
    no_comment_on_failure: bool,
    dry_run: bool,
    debug: bool,
    format: str,
) -> None:
    """Verifies issues in Jira."""
    issues: List[Issue] = []
    verifeid_issues: List[Issue] = []
    jiav_logger: Logger = logger.configure_logger(debug)
    # If user requested to upload attachment, warn them
    if upload_attachment:
        jiav_logger.warn(
            "".join(
                [
                    "Will upload attachment of the execution, this might leak ",
                    "sensetive information, "
                    "refer to: https://jira.atlassian.com/browse/JRASERVER-3893",
                ]
            )
        )
    # If user requested to include public comments, warn them
    if allow_public_comments:
        jiav_logger.warn("Will include public comments in search!")
    if no_comment_on_failure:
        jiav_logger.warn("Comment will not be posted of failed manifest execution")
    if dry_run:
        jiav_logger.info("Will not update issues, running as a dry run")
    try:
        jira_connection: JiraConnection = JiraConnection(
            url=jira,
            username=username,
            access_token=access_token,
        )
    except exceptions.NoJiraRestAPIEndpoint as e:
        jiav_logger.exception(e)
        jiav_logger.critical("Provided URL does not contain a valid JIRA endpoint")
        raise SystemExit(1)
    except exceptions.JiraUnhandledException as e:
        jiav_logger.exception(e)
        jiav_logger.critical(
            "An unhandled exception occurred while trying to connect to Jira"
        )
        raise SystemExit(1)
    except exceptions.JiraAuthenticationFailed as e:
        jiav_logger.exception(e)
        jiav_logger.critical(
            " ".join(
                [
                    "Authentication with the Jira insatance failed,",
                    "please check your credentials",
                ]
            )
        )
        raise SystemExit(2)
    except exceptions.JiraMissingCredentials as e:
        jiav_logger.exception(e)
        jiav_logger.critical(
            "Missing username, please provide a username to a Jira cloud instance"
        )
        raise SystemExit(2)
    # Fetch issues from authenticated Jira instance
    try:
        issues = jira_connection.fetch_issues(issues=issue, jql=query)
    except exceptions.JiraMissingCredentials:
        pass
    except exceptions.JiraUnhandledException as e:
        jiav_logger.exception(e)
        jiav_logger.critical(
            "An unhandled exception occurred while trying to fetch issues"
        )
        raise SystemExit(2)
    except exceptions.InvalidKeyInJQL:
        jiav_logger.critical("Invalid key in JQL")
        raise SystemExit(3)
    except exceptions.JQLReturnedNothing as e:
        jiav_logger.exception(e)
        jiav_logger.critical("Query returned no issues")
        raise SystemExit(3)
    except exceptions.JQLError as e:
        jiav_logger.exception(e)
        jiav_logger.critical("Query returned an error")
        raise SystemExit(3)
    except exceptions.NoIssuesFound:
        jiav_logger.critical("No issues found")
        raise SystemExit(4)
    # Iterate over valid issues and attempt to update and verify them
    verifeid_issues = verification.verify_issues(
        issues=issues,
        jira_connection=jira_connection,
        upload_attachment=upload_attachment,
        allow_public_comments=allow_public_comments,
        no_comment_on_failure=no_comment_on_failure,
        dry_run=dry_run,
    )
    # Print summary if issues were verified
    if verifeid_issues:
        summary.prepare_summary(issues=verifeid_issues, format=format)
    else:
        jiav_logger.info("No issues were verified")
        raise SystemExit(6)


jiav.add_command(validate_manifest)
jiav.add_command(verify)

if __name__ == "__main__":
    jiav()



#!/usr/bin/env python

import tempfile
from datetime import datetime
from typing import IO, Any, List

from iteration_utilities import deepflatten  # type:ignore


def get_current_timestamp() -> str:
    """
    Get current timestamp

    Returns:
        current_timestamp - A formatted string of current time:
                            <month><day><year><hour><minutes><seconds>
    """
    current_timestamp = datetime.now().strftime("%m%d%Y%H%M%S")
    return current_timestamp


def write_content_to_tempfile(content: List[Any]) -> IO:
    """
    Write content to temporary file

    Arguments:
        content - Content to write

    Returns:
        temp_file - temporary file
    """
    # If we receive an output that contains a list
    # flatten it to a single list
    if isinstance(content, list):
        content = list(deepflatten(content, depth=1))
    # Create temporary file
    temp_file = tempfile.NamedTemporaryFile(
        prefix=f"jiav_{get_current_timestamp()}_", suffix=".txt", mode="w+t"
    )
    # Write content to temporary file
    temp_file.writelines(content)
    # Commit changes to file
    temp_file.seek(0)
    return temp_file



#!/usr/bin/env python

import logging
from logging import Logger

from rich.console import Console
from rich.logging import RichHandler


def subscribe_to_logger() -> Logger:
    """
    Subscribes module to general logger

    Returns:
        logger - Logger object for that module
    """
    logger = logging.getLogger(__name__)
    return logger


def configure_logger(debug: bool = False) -> Logger:
    """
    Initializes a logger handler to log events

    Arguments:
        debug - Enable debug mode during invocation

    Returns:
        logger - Logger handler
    """
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                omit_repeated_times=False,
                console=Console(stderr=True),
            )
        ],
    )
    logger = subscribe_to_logger()
    if debug:
        logger.setLevel(logging.DEBUG)
    return logger
